!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: setpathterm.gs,v 1.8 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy:
!   SetValuedPathTerm, PathTerm, Array, SequenceableCollection, Collection,
!   Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods SetValuedPathTerm
removeallclassmethods SetValuedPathTerm
protectmethods

category: 'For Documentation Installation only'
classmethod: SetValuedPathTerm
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Instance methods for SetValuedPathTerm
category: 'Updating Indexes'
method: SetValuedPathTerm
addMappingsUsing: mapInfo logging: aBoolean

"Adds index dictionary entries for the objects in the NSC described by the
 mapInfo.  Also update the NSC's index list."
<protected>

| iList anNsc nextPathTerm indexDict setElementMap nextMaps dictUpdated |
anNsc := mapInfo object.

( nil == anNsc _or: [ self size == 0 ] )
    ifTrue: [ ^ self ].

dictUpdated := false.
iList := anNsc _getIndexList.
anNsc _putInWriteSet.
" for each index that utilizes this path term "
1 to: self size do: [ :i |
    iList addIndex: (self at: i) withOffset: offset + 1 nsc: anNsc
].

indexDict := self getIndexDictionary.
" for each child mapping "
1 to: mapInfo size do: [ :k |
    nextMaps := mapInfo at: k.
    " there is a collection of mappings for each element in the set "
    1 to: nextMaps size do: [ :i |
        setElementMap := nextMaps at: i.

        dictUpdated
            ifFalse: [
                " insert dictionary entry mapping set element -> NSC "
                indexDict
                    _at: setElementMap object
                    put: anNsc
                    term: self
                    logging: aBoolean.

                self addSegment: setElementMap object segment logging: aBoolean
            ].

        updateBtree ~~ nil
            ifTrue: [
                self addDirectMappingFor: setElementMap object logging: aBoolean
            ].

        (nextPathTerm := setElementMap pathTerm) ~~ nil
            ifTrue: [
                nextPathTerm addMappingsUsing: setElementMap logging: aBoolean
            ]
    ].
    dictUpdated := true
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
getMappingInfoFor: anNsc
ifObject: anObject
atOffset: ivOffset
replaceWith: newValue

"Gathers all the objects, their dependency lists, and instance variable offset
 for each object in anNsc and place these in a MappingInfo object.  Returns the
 mapping info."
<protected>

| mapInfo maps iList |

( nil == anNsc _or: [ self size == 0 ] )
    ifTrue: [ ^ MappingInfo new pathTerm: self ].

anNsc class isNsc
    ifFalse: [
        ^ #[ false, " indicating no indexing objects were modified "
            (ErrorSymbols at: #rtErrPathTermObjectNotAnNsc),
            anNsc,
            nil ]
    ].

iList := anNsc _indexedPaths.
iList ~~ nil
  ifTrue: [
    " for each index on the NSC ... "
    1 to: iList size by: 2 do: [ :i |
      " for each index in which this path term is used ... "
      1 to: self size do: [ :j |
        ( (iList at: i) == (self at: j) _and:
        [ (iList at: i + 1) ~= (offset + 1) ] )
          ifTrue: [
            ^ #[ false, " indicating no indexing objects were modified "
                (ErrorSymbols at: #rtErrNscParticipatesInMultipleTerms),
                anNsc,
                (iList at: i) pathComponentsString ]
          ]
      ]
    ]
  ].

" create a mapping info object for the NSC "
mapInfo := MappingInfo new object: anNsc; pathTerm: self.

updateBtree ~~ nil
    ifTrue: [
        maps := Array new.
        anNsc do: [ :setElement |
            " the set elements mapping info contains just the set element "
            maps add: (MappingInfo new object: setElement)
        ].
        mapInfo addLast: maps.
        ^ mapInfo
    ].

children isEmpty
    ifTrue: [ 
        maps := Array new.
        anNsc do: [ :setElement |
            maps add: (MappingInfo new object: setElement)
        ].
        mapInfo addLast: maps
    ]
    ifFalse: [
        " for each following path term "
        1 to: children size do: [ :i | 
            maps := Array new.
            anNsc do: [ :setElement | | aMapInfo |
                " get the set elements mapping info and add it to the set "
                aMapInfo := (children at: i)
                    getMappingInfoFor: setElement
                    ifObject: anObject
                    atOffset: ivOffset
                    replaceWith: newValue.

                (aMapInfo _class == MappingInfo)
                    ifFalse: [ ^ aMapInfo ].
                maps add: aMapInfo
            ].
            mapInfo addLast: maps
        ]
    ].

^ mapInfo
%

category: 'Testing'
method: SetValuedPathTerm
indicatesIndexOnNscElements

"Returns true if the path term indicates that the index is on elements of the
 NSC itself.  This is the case if the path term ends in '*' (that is, this
 path term has no children)."
<protected>

^ children isEmpty
%

category: 'Testing'
method: SetValuedPathTerm
indicatesNsc

"Returns true, the receiver indicates that indexing is on a set-valued instance
 variable."
<unprotected>

^ true
%

category: 'Updating Indexes'
method: SetValuedPathTerm
removeMappingsFor: anNsc lastOne: aBoolean logging: doLogging

"Removes entries in the index dictionary, index list, and dependency lists for
 the elements contained in the NSC."
<protected>

| iList indexDict sz |
self size == 0
  ifTrue: [ ^ self ].

aBoolean
  ifTrue: [
    (iList := anNsc _indexedPaths) ~~ nil
      ifTrue: [ iList removeIndexesInPathTerm: self for: anNsc ]
  ].

indexDict := self getIndexDictionary.
sz := children size.
" for each object in the NSC "
anNsc do: [ :setElement | | more depList lastOne |
  " remove the index dictionary entry "
  more := indexDict removeKey: setElement value: anNsc term: self logging: doLogging.
  lastOne := aBoolean _and: [ more not ].

  updateBtree ~~ nil
    ifTrue: [
      self removeDirectMappingFor: setElement logging: doLogging.
      needsDepList
        ifTrue: [ DependencyList removePathTerm: self for: setElement ]
    ].

  nil ~~ setElement
    ifTrue: [
      " make recursive call to remove mappings "
      1 to: sz do: [ :j |
        (children at: j) removeMappingsFor: setElement
          lastOne: lastOne
          logging: doLogging
      ]
    ].
  lastOne
    ifTrue: [
      (depList := DependencyList for: setElement) ~~ nil
        ifTrue: [ depList removePathTerms: children for: setElement ]
    ]
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
addMappingsForObject: anNsc logging: aBoolean

"Adds index dictionary entries for the objects in anNsc.  Also update the
 NSC's index list."
<protected>

| iList indexDict sz |

( nil == anNsc _or: [ self size == 0 ] )
    ifTrue: [ ^ self ].

anNsc class isNsc
    ifFalse: [  ^ self _errorPathObjectNotAnNsc: anNsc ].

sz := self size.
iList := anNsc _getIndexList.
anNsc _putInWriteSet.
" for each index that utilizes this path term "
1 to: sz do: [ :i |
    iList addIndex: (self at: i) withOffset: offset + 1 nsc: anNsc
].

sz := children size.
indexDict := self getIndexDictionary.
anNsc do: [ :setElement |
    " insert dictionary entry mapping set element -> NSC "
    indexDict _at: setElement put: anNsc term: self logging: aBoolean.

    self addSegment: setElement segment logging: aBoolean.

    updateBtree ~~ nil
        ifTrue: [ self addDirectMappingFor: setElement logging: false ].

    1 to: sz do: [ :j |
        (children at: j) addMappingsForObject: setElement
            logging: aBoolean
    ]
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
addIndexListEntriesFor: nscBag index: indexObj

"Update the Bag of NSC's index lists."
<protected>

| anNsc iList nextOffset |

nextOffset := offset + 1.
" for each NSC in the Bag of NSCs ... "
1 to: nscBag size do: [ :i |
    anNsc := nscBag at: i.
    anNsc class isNsc
        ifFalse: [  ^ self _errorPathObjectNotAnNsc: anNsc ].

    iList := anNsc _getIndexList.
    anNsc _putInWriteSet.

    " update the index list "
    iList addIndex: indexObj withOffset: nextOffset nsc: anNsc.

    " add the NSC's segment to the segment set "
    self addSegment: anNsc segment logging: false
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
removeIndexListEntriesFor: nscBag index: indexObj

"Update the Bag of NSC's index lists."
<protected>

| anNsc iList nextOffset |

nextOffset := offset + 1.
" for each NSC in the Bag of NSCs ... "
1 to: nscBag size do: [ :i |
    anNsc := nscBag at: i.
    iList := anNsc _getIndexList.
    anNsc _putInWriteSet.
    " update the index list "
    iList removeIndex: indexObj
        withOffset: nextOffset
        for: anNsc

]
%

category: 'Audit'
method: SetValuedPathTerm
auditObject: anNsc occurrences: num on: aString

"Private.  Unsupported method for Technical Support."

"Verify the given object has the correct dependency list, index
 dictionary, and B-tree entries for the receiver. "
<protected>

| iList indexObj j sz |
( nil == anNsc _or: [ self size == 0 ] )
    ifTrue: [ ^ aString ].

anNsc class isNsc
    ifFalse: [
        self auditInfo: #[ anNsc, nil, nil ] on: aString.
        ^ aString add: 'Object is not an NSC'; add: Character lf
    ].

sz := self size.
iList := anNsc _indexedPaths.
iList == nil
  ifTrue: [
    self auditInfo: #[ anNsc, nil, nil ] on: aString.
    aString add: 'NSC does not contain index list when it should';
      add: Character lf
  ]
  ifFalse: [
    " for each index that utilizes this path term "
    1 to: sz do: [ :i |
        indexObj := self at: i.
        j := iList _findOffsetFor: indexObj withOffset: offset + 1.
        j == nil
            ifTrue: [
                self auditInfo: #[ anNsc, indexObj, nil ] on: aString.
                aString add: 'NSC does not contain index list entry for index';
                    add: Character lf
            ]
    ]
  ].

anNsc do: [ :setElement | | occurs k |
    occurs := anNsc occurrencesOf: setElement.
    k := self auditDictionaryEntryFor: anNsc
        and: setElement
        occurrences: num * occurs
        on: aString.

    nil ~~ setElement
        ifTrue: [
            children isEmpty
                ifTrue: [
                    self auditDirectObject: setElement
                        occurrences: k
                        on: aString
                ]
                ifFalse: [
                    1 to: children size do: [ :j |
                        (children at: j) auditObject: setElement
                            occurrences: k
                            on: aString
                    ]
                ]
        ]
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
cleanupDependencyListFor: anNsc

""
<protected>

| sz iList |

( nil == anNsc _or: [ self size == 0 ] )
  ifTrue: [ ^ self ].

anNsc class isNsc
  ifFalse: [  ^ self _errorPathObjectNotAnNsc: anNsc ].

(iList := anNsc _indexedPaths) ~~ nil
  ifTrue: [ iList removeIndexesInPathTerm: self for: anNsc ].

sz := children size.
anNsc do: [ :setElement | | depList |
  needsDepList
    ifTrue: [
      depList := DependencyList for: setElement.
      depList ~~ nil
        ifTrue: [ depList removeCompletelyPathTerm: self for: setElement ]
    ].

  1 to: sz do: [ :j |
    (children at: j) cleanupDependencyListFor: setElement
  ]
]
%

category: 'Updating Indexes'
method: SetValuedPathTerm
_pretraverse: anNsc
ifObject: anObject
atOffset: ivOffset
replaceWith: newValue

"Performs a pretraversal along the index path.  If there is an error while
 traversing, returns an Array of error information.  Otherwise returns self."
<protected>

| iList |

( nil == anNsc _or: [ self size == 0 ] )
    ifTrue: [ ^ self ].

anNsc class isNsc
  ifFalse: [
    ^ #[ false, " indicating no indexing objects were modified "
        (ErrorSymbols at: #rtErrPathTermObjectNotAnNsc),
        anNsc,
        nil ]
  ].

iList := anNsc _indexedPaths.
iList ~~ nil
  ifTrue: [
    " for each index on the NSC ... "
    1 to: iList size by: 2 do: [ :i |
      " for each index in which this path term is used ... "
      1 to: self size do: [ :j |
        ( (iList at: i) == (self at: j) _and:
        [ (iList at: i + 1) ~= (offset + 1) ] )
          ifTrue: [
            ^ #[ false, " indicating no indexing objects were modified "
                (ErrorSymbols at: #rtErrNscParticipatesInMultipleTerms),
                anNsc,
                (iList at: i) pathComponentsString ]
          ]
      ]
    ]
  ].

children isEmpty
  ifFalse: [
    " for each following path term "
    1 to: children size do: [ :i |
      anNsc do: [ :setElement | | result |
        " get the set elements mapping info and add it to the set "
        result := (children at: i)
          _pretraverse: setElement
          ifObject: anObject
          atOffset: ivOffset
          replaceWith: newValue.

        result == (children at: i)
          ifFalse: [ ^ result ].
      ].
    ]
  ].

^ self
%

category: 'Auditing'
method: SetValuedPathTerm
allDependencyListsFor: anNsc into: all

"Put any dependency lists that can be found from the receiver
into the given set."

<protected>

| sz |

( nil == anNsc _or: [ self size == 0 ] )
  ifTrue: [ ^ self ].

anNsc class isNsc
  ifFalse: [  ^ self _errorPathObjectNotAnNsc: anNsc ].

sz := children size.
anNsc do:
  [:setElement | | depList |
  needsDepList
    ifTrue: [
      depList := DependencyList for: setElement.
      depList ~~ nil
        ifTrue: [ all add: depList ].
    ].

  1 to: sz do: [ :j |
    (children at: j) allDependencyListsFor: setElement into: all
  ]
]
%

unprotectmethods
